from tef import TEF6686
from protocol import I2CPCClient

p = I2CPCClient("COM17")
tef = TEF6686(p)
tef.init()
tef.FM_Tune_To(1, 9500)
tef.AUDIO_Set_Mute(False)
tef.AUDIO_Set_Volume(70)
tef.FM_Set_MphSuppression(True)
tef.FM_Set_ChannelEqualizer(True)
tef.FM_Set_Specials(1)
tef.FM_Set_Bandwidth(True)
import time
time.sleep(0.032)
try:
    while True:
        status, level, usn, wam, offset, bandwidth, modulation = tef.FM_Get_Quality_Data()
        if not status or not level or not usn or not wam or not offset or not bandwidth or not modulation: continue
        print(f"{level / 10} dbμV | {usn / 10}% USN | {wam / 10}% WAM | {offset / 10} kHz offset | {bandwidth / 10} kHz bandwidth | {modulation / 10}% modulation")
        time.sleep(0.3)
finally:
    tef.close()